#ifndef PEW_INTRINSIC
#define PEW_INTRINSIC

#include <cstdint>
#include <emmintrin.h>
#include <immintrin.h>
#include <smmintrin.h>
#include <sstream>
#include <tmmintrin.h>
#include <xmmintrin.h>

namespace pew {
    using imm = uint8_t;
    template<typename T>
    struct intrinsic;

    template<>
    struct intrinsic<double> {
        using type                    = __m256d;
        static const size_t alignment = 32;

        inline static auto
        set(double x, double y, double z) {
            return _mm256_set_pd(20, z, y, x);
        }

        inline static auto
        set(double c) {
            return _mm256_set1_pd(c);
        }

        inline static auto
        store(double* arr, type vec) {
            _mm256_store_pd(arr, vec);
            return arr;
        }

        inline static auto
        mul(const type& left, const type& right) {
            return _mm256_mul_pd(left, right);
        }

        inline static auto
        add(const type& left, const type& right) {
            return _mm256_add_pd(left, right);
        }

        inline static auto
        sub(const type& left, const type& right) {
            return _mm256_sub_pd(left, right);
        }

        inline static auto
        div(const type& left, const type& right) {
            return _mm256_div_pd(left, right);
        }

        inline static auto
        div(const type& vec, double s) {
            return div(vec, set(s));
        }

        inline static auto
        mul(const type& vec, double s) {
            return mul(vec, set(s));
        }

        inline static auto
        dot(const type& left, const type& right) -> type {
            auto reg_xy = mul(left, right);

            // clang-format off
            return _mm256_permute2f128_pd(
                _mm256_permute_pd(
                    add(
                        _mm256_hadd_pd(reg_xy, reg_xy),
                        _mm256_permute2f128_pd(reg_xy, reg_xy, 0x81)
                    ), 0),
                left, 0);
            // clang-format on
        }

        inline static auto
        sqrt(const type& reg) {
            return _mm256_sqrt_pd(reg);
        }

        inline static auto
        len(const type& reg) {
            return sqrt(dot(reg, reg));
        }

        inline static auto
        unit(const type& reg) {
            return div(reg, len(reg));
        }

        inline static auto
        sign(const type& vec) {
            return mul(vec, set(-1.0));
        }

        inline static auto
        cmplt(const type& left, const type& right) -> bool {
            auto reg_ge = _mm256_cmp_pd(left, right, _CMP_GE_OQ);
            return _mm256_testz_pd(reg_ge, reg_ge);
        }

        inline static auto
        abs(const type& abs) {
            return _mm256_blendv_pd(abs, sign(abs), _mm256_cmp_pd(abs, set(0), _CMP_LT_OQ));
        }

        inline static auto
        dump(const type& reg) {
            alignas(alignment) double arr[4];
            store(arr, reg);
            std::stringstream stream;
            stream << "[" << arr[0] << " " << arr[1] << " " << arr[2] << " " << arr[3] << "]";
            return stream.str();
        }
    };

    template<>
    struct intrinsic<int> {
        using type                         = __m128i;
        static constexpr size_t alignment  = 8;
        static constexpr auto   store_mask = 16;

        inline static auto
        set(int x, int y, int z) {
            return _mm_set_epi32(0, z, y, x);
        }

        inline static auto
        set(int c) {
            return _mm_set1_epi32(c);
        }

        inline static auto
        store(int* ptr, const type& vec) {
            return _mm_store_epi32(ptr, vec);
        }

        inline static auto
        mul(const type& left, const type& right) {
            return _mm_mullo_epi32(left, right);
        }

        inline static auto
        mul(const type& vec, int s) {
            return mul(vec, set(s));
        }

        inline static auto
        sign(const type& vec) {
            return _mm_sign_epi32(vec, set(-1));
        }

        inline static auto
        add(const type& left, const type& right) {
            return _mm_add_epi32(left, right);
        }

        inline static auto
        sub(const type& left, const type& right) {
            return _mm_sub_epi32(left, right);
        }

        inline static int
        dot(const type& left, const type& right) {
            alignas(alignment) int arr[4];
            store(arr, mul(left, right));
            return arr[0] + arr[1] + arr[2];
        }

        inline static auto
        div(const type& vec, int s) {
            alignas(alignment) int arr[4];
            store(arr, vec);
            return set(arr[0] / s, arr[1] / s, arr[2] / s);
        }

        inline static auto
        abs(const type& abs) {
            return _mm_abs_epi32(abs);
        }
    };

    template<>
    struct intrinsic<float> {
        using type                    = __m128;
        static const size_t alignment = 32;

        inline static auto
        set(float x, float y, float z) {
            return _mm_set_ps(0, z, y, x);
        }

        inline static auto
        set(float c) {
            return _mm_set1_ps(c);
        }

        inline static auto
        store(float* ptr, type vec) {
            return _mm_store_ps(ptr, vec);
        }

        inline static auto
        mul(const type& left, const type& right) {
            return _mm_mul_ps(left, right);
        }

        inline static auto
        mul(const type& vec, float s) {
            return mul(vec, set(s));
        }

        inline static auto
        sign(const type& vec) {
            return mul(vec, set(-1.0));
        }

        inline static auto
        add(const type& left, const type& right) {
            return _mm_add_ps(left, right);
        }

        inline static auto
        sub(const type& left, const type& right) {
            return _mm_sub_ps(left, right);
        }

        template<imm mask = 0xff>
        inline static auto
        dot(const type& left, const type& right) {
            return _mm_dp_ps(left, right, mask);
        }

        inline static auto
        div(const type& vec, float s) {
            return _mm_div_ps(vec, set(s));
        }

        inline static auto
        abs(const type& reg) {
            auto mask = _mm_cmplt_ps(reg, set(0));
            return _mm_blendv_ps(reg, sign(reg), mask);
        }
    };
} // namespace pew

#endif